/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.core;

import org.apache.muse.ws.addressing.soap.SoapFault;

/**
 * 
 * Persistence is a mechanism used by resources and/or capabilities to save 
 * some or all of their state across application reboots. Concrete persistence 
 * classes provide the means for storage (file, database, etc.) as well as 
 * the logic to determine what content is stored. Note that this interface 
 * only describes how the persistence class should reload its data - not how 
 * it should initially find and save this data.
 * <br><br>
 * Peristence implementations are usually specified with the <em>persistence</em>
 * element in muse.xml. This element allows users to specify the persistence 
 * location (the semantics of which are implementation-specific) and the concrete 
 * class that implements this interface.
 *
 * @author Dan Jemiolo (danj)
 *
 */
public interface Persistence extends InitializationParameters
{
    /**
     * 
     * @return The value specified in muse.xml under the <em>persistence</em> 
     *         element. The semantics of this value are dependent on the type 
     *         of storage used by the implementation class. As an example, a 
     *         simple file-based persistence class might interpret the location 
     *         to be the directory where files are stored; a database-oriented 
     *         class might interpret it as the JNDI URI for the database.
     *
     */
    String getPersistenceLocation();
    
    /**
     * 
     * @return The ResourceManager that stores all of the resource type 
     *         definitions and all current resource instances.
     *
     */
    ResourceManager getResourceManager();
    
    /**
     * 
     * Reads any data that was persisted by the class back into memory and 
     * provides it to the proper components (resource or capability). This 
     * method is normally called during the initialization of said components, 
     * before any requests are handled by the application.
     *
     * @throws SoapFault
     *
     */
    void reload() 
        throws SoapFault;
    
    void setPersistenceLocation(String location);
    
    void setResourceManager(ResourceManager manager);
}
