/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.core.proxy;

import javax.xml.namespace.QName;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;

/**
 *
 * AbstractProxyHandler is a convenience class that implements all of the 
 * accessor methods of {@linkplain ProxyHandler ProxyHandler}, leaving only 
 * the serialization methods (fromXML(Element) and toXML(Object[])).
 *
 * @author Dan Jemiolo (danj)
 *
 */

public abstract class AbstractProxyHandler implements ProxyHandler
{    
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = 
        MessagesFactory.get(AbstractProxyHandler.class);
    
    //
    // The WS-A Action URI - can be used for building SOAP messages or 
    // uniquely identifying the handler
    //
    private String _action = null;
        
    //
    // The names of the individual parameters (inside the request element)
    //
    private QName[] _requestParamQNames = null;
    
    //
    // The names of the individual parameter XSD types
    //
    private QName[] _requestParamSchemaTypes = null;
    
    //
    // The Java types of the individual parameters (or an empty array 
    // if not available)
    //
    private Class[] _requestParamTypes = new Class[0];
    
    //
    // The name of the request message (in the SOAP Body)
    //
    private QName _requestQName = null;
    
    //
    // The name of the response message (in the SOAP Body). If this is 
    // null, the handler accepts "xsd:anyType"
    //
    private QName _responseQName = null;
    
    //
    // The type of the response message (in the SOAP Body). If this is 
    // null, the handler accepts "xsd:anyType"
    //
    private QName _returnSchemaType = null;
    
    //
    // The return type used to deserialize the response
    //
    private Class _returnType = null;
    
    public String getAction()
    {
        return _action;
    }
        
    public QName getRequestName()
    {
        return _requestQName;
    }

    public QName[] getRequestParameterNames()
    {
        return _requestParamQNames;
    }

    public QName[] getRequestParameterSchemaTypes()
    {
        return _requestParamSchemaTypes;
    }

    public Class[] getRequestParameterTypes()
    {
        return _requestParamTypes;
    }
    
    public QName getResponseName()
    {
        return _responseQName;
    }
    
    public QName getReturnSchemaType()
    {
        return _returnSchemaType;
    }
    
    public Class getReturnType()
    {
        return _returnType;
    }
    
    public void setAction(String action)
    {
        if (action == null)
            throw new NullPointerException(_MESSAGES.get("NullAction"));
        
        _action = action;
    }
    
    public void setRequestName(QName requestQName)
    {
        if (requestQName == null)
            throw new NullPointerException(_MESSAGES.get("NullRequestName"));
        
        _requestQName = requestQName;
    }
    
    public void setRequestParameterNames(QName[] parameterQNames)
    {
        if (parameterQNames == null)
            throw new NullPointerException(_MESSAGES.get("NullRequestParams"));
        
        _requestParamQNames = parameterQNames;
    }
    
    public void setRequestParameterSchemaTypes(QName[] parameterTypes)
    {
        if (parameterTypes == null)
            parameterTypes = new QName[0];
        
        _requestParamSchemaTypes = parameterTypes;
    }
    
    public void setRequestParameterTypes(Class[] parameterTypes)
    {
        if (parameterTypes == null)
            parameterTypes = new Class[0];
        
        _requestParamTypes = parameterTypes;
    }
    
    public void setResponseName(QName responseQName)
    {
        _responseQName = responseQName;
    }
    
    public void setReturnSchemaType(QName returnType)
    {
        _returnSchemaType = returnType;
    }
    
    public void setReturnType(Class returnType)
    {
        //
        // accept both "null" and "void" for void methods
        //
        _returnType = returnType == null ? Void.TYPE : returnType;
    }
}
