/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/


package org.apache.muse.core.platform.osgi;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceListener;

import org.apache.muse.core.platform.osgi.routing.OSGiResourceRouter;
import org.apache.muse.core.platform.osgi.util.OSGiReflectUtilHelper;

/**
*
* Activator is the <code>BundleActivator</code> implementation for the Muse Core 
* Bundle. The activator manages the OSGi Service interface to the appropriate IsolationLayer 
* implementation through an instance of the 
* <code>org.apache.muse.core.platform.osgi.ResourceManagementProvider</code> service.
* 
* @see org.apache.muse.core.platform.osgi.ResourceManagementProvider
* @author Joel Hawkins (joelh)
*
*/

public class Activator implements BundleActivator {

	private static final String MANAGEMENT_ENVIRONMENT_FILTER = "(objectclass=org.apache.muse.core.platform.osgi.ResourceManagementProvider)";

	private ServiceReference managementServiceRef;

	private static Activator activator;

	private BundleContext context;

	private ResourceManagementProvider provider;

	static private class ManagementServiceListener implements ServiceListener {

		public void serviceChanged(ServiceEvent event) {
			ServiceReference ref = event.getServiceReference();
			switch (event.getType()) {
			case ServiceEvent.REGISTERED:
				if (activator.managementServiceRef == null) {
					activator.managementServiceRef = event.getServiceReference();
					activator.provider = (ResourceManagementProvider) activator.context
							.getService(activator.managementServiceRef);
					OSGiResourceRouter.getDefault().attachManagementProvider(
							activator.provider);
				}
				break;
			case ServiceEvent.MODIFIED:
				break;
			case ServiceEvent.UNREGISTERING:
				ref = event.getServiceReference();
				if (ref == activator.managementServiceRef) {
					activator.managementServiceRef = null;
					OSGiResourceRouter.getDefault().releaseManagementProvider();
				}
				break;
			}
		}
	}

	private ManagementServiceListener managementListener = new ManagementServiceListener();

	/**
	 * starts the Bundle. The <code>start</code> method attempts to resolve an initial
	 * refrence to the <code>ResourceManagementProvider</code> service and attach that 
	 * service to the <code>OSGiResourceRouter</code> instance. The method also intializes
	 * the OSGiReflectUtilHelper instance.
	 *
	 * @see org.apache.muse.core.platform.osgi.routing.OSGiResourceRouter
	 * @see org.apache.muse.core.platform.osgi.util.OSGiReflectUtilHelper
	 * @see org.osgi.framework.BundleActivator#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		activator = this;
		this.context = context;
		try {
			OSGiResourceRouter.getDefault().initialize(context);
		} catch (Throwable t) {
			t.printStackTrace();
		}

		if ((managementServiceRef = context.getServiceReference("org.apache.muse.core.platform.osgi.ResourceManagementProvider")) != null) {
			activator.provider = (ResourceManagementProvider) activator.context.getService(activator.managementServiceRef);
			OSGiResourceRouter.getDefault().attachManagementProvider(activator.provider);
		}
		
		OSGiReflectUtilHelper.getDefault().initialize(context);

		context.addServiceListener(managementListener, MANAGEMENT_ENVIRONMENT_FILTER);

	}

	/**
	 * stops the Bundle and removes any applied listeners
	 *
	 * @see org.osgi.framework.BundleActivator#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
		context.removeBundleListener(OSGiResourceRouter.getDefault());
		context.removeServiceListener(managementListener);
	}

}
