/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/


package org.apache.muse.core.platform.osgi.descriptor;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.io.File;
import java.io.InputStream;

import javax.wsdl.Definition;
import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.Part;
import javax.wsdl.PortType;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;
import javax.wsdl.xml.WSDLLocator;

import org.apache.muse.core.Environment;
import org.apache.muse.core.Resource;
import org.apache.muse.core.descriptor.CapabilityDescriptor;
import org.apache.muse.core.descriptor.DescriptorConstants;
import org.apache.muse.core.descriptor.ResourceDefinition;
import org.apache.muse.core.descriptor.SimpleResourceDescriptor;
import org.apache.muse.core.descriptor.WsdlConfig;
import org.apache.muse.util.ReflectUtils;
import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.util.xml.XsdUtils;
import org.apache.muse.ws.wsdl.WsdlUtils;
import org.osgi.framework.Bundle;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

import org.apache.muse.core.platform.osgi.util.OSGiReflectUtilHelper;

import java.net.URL;

/**
 *
 * OSGiResourceDescriptor is an OSGi-aware implementation of the 
 * <code>SimpleResourceDescriptor</code> class, which implements the 
 * <code>ResourceDescriptor</code> interface
 * 
 * @see ResourceDecriptor
 * @author Joel Hawkins (joelh)
 *
 */

public class OSGiResourceDescriptor extends SimpleResourceDescriptor {

	private static class OSGiWsdlLocator implements WSDLLocator {

		private Bundle bundle;

		private WsdlConfig wsdlConfig;

		private String baseURI = "";

		private String lastPath = "";

		public OSGiWsdlLocator(Bundle bundle, WsdlConfig wsdlConfig) {
			this.bundle = bundle;
			this.wsdlConfig = wsdlConfig;
		}

		public InputSource getBaseInputSource() {
			String path = wsdlConfig.getWsdlPath();
			URL url = bundle.getResource(path);
			InputStream stream = null;
			if(url == null){
				stream = OSGiReflectUtilHelper.getDefault().getResource(path);
			}
			try {
				if(stream == null) stream = url.openStream();
			} catch (Throwable t) {
				return null;
			}
			if (stream != null)
				return new InputSource(stream);
			return null;
		}

		public String getBaseURI() {
			// TODO clean up this code
			if (wsdlConfig.getWsdlPath().startsWith("/OSGI-INF/wsdl"))
				baseURI = "/OSGI-INF/wsdl";
			if(wsdlConfig.getWsdlPath().startsWith("/wsdl"))
				baseURI = "/wsdl";
			baseURI="";
			return baseURI;
		}

		public InputSource getImportInputSource(String base, String path) {
			lastPath = path;
			String fullPath = path;
			if (base != null) {
				fullPath = base + "/" + path;
			}
			URL url = bundle.getResource(fullPath);
			InputStream stream = null;
			if(url == null){
				stream = OSGiReflectUtilHelper.getDefault().getResource(fullPath);
			}
			try {
				if(stream == null) stream = url.openStream();
				this.lastPath = fullPath;
			} catch (Throwable t) {
				return null;
			}
			if (stream != null)
				return new InputSource(stream);
			return null;
		}

		public String getLatestImportURI() {
			return lastPath;
		}

	};

	private static Messages _MESSAGES = MessagesFactory
			.get(SimpleResourceDescriptor.class);

	protected Bundle bundle;

	public void setBundle(Bundle bundle) {
		this.bundle = bundle;
	}

	/**
	 * resolves the resource class from the resource decriptor
	 * @param xml the xml Element representing the resource class
	 * @param env the Muse Environment
	 * @return the Class for the resource
	 */
	protected Class createResourceClass(Element xml, Environment env) {
		QName qname = DescriptorConstants.JAVA_RESOURCE_QNAME;
		String className = XmlUtils.getElementText(xml, qname);

		if (className == null)
			throw new RuntimeException(_MESSAGES.get("NullJavaBase"));
		ClassLoader loader = env.getClassLoader();
		Class resourceClass = null;

		if (bundle == null) {
			loader = env.getClassLoader();
			resourceClass = ReflectUtils.getClass(className, loader);
		} else {
			try {
				resourceClass = ReflectUtils.getClass(className,loader);
				if(resourceClass == null){
					resourceClass = bundle.loadClass(className);
				}
			} catch (ClassNotFoundException e) {
				throw new RuntimeException("Class Not Found " + className);
			}
		}

		if (!Resource.class.isAssignableFrom(resourceClass)) {
			Object[] filler = { className, Resource.class.getName() };
			String message = _MESSAGES.get("IncorrectResourceRoot", filler);
			throw new RuntimeException(message);
		}

		return resourceClass;
	}


	/**
	 * returns the operations from the wsdl document, using an internal WSDLLocator implementation
	 * that allows access to wsdl and schema imports from external Bundles.
	 * @param wsdlDoc the <code>Document</code> representing the wsdl for the resource type
	 * @param wsdlConfig the wsdl configuration used to establish context for other documents
	 *                   referenced from the inital wsdl.
	 * @param env the Muse Environment
	 * @return a <code>Map</code> containing the wsdl Operations keyed by their respective 
	 *         WSA Actions                  
	 */
	protected Map getWsdlOperations(Document wsdlDoc,
			WsdlConfig wsdlConfig,
			Environment env)
	{
		String wsdlPath = wsdlConfig.getWsdlPath();
		String absolutePath = env.getRealDirectory().getAbsolutePath();

		if (wsdlPath.charAt(0) != '/')
			absolutePath += '/';

		absolutePath += wsdlPath;
		File wsdlFile = new File(absolutePath);
		File wsdlDir = wsdlFile.getParentFile();

		Definition def = null;

		//
		// use WSDL4J to parse/validate the WSDL
		//
		try {

			WSDLFactory factory = WSDLFactory.newInstance();
			WSDLReader reader = factory.newWSDLReader();
			javax.wsdl.xml.WSDLLocator locator = new OSGiWsdlLocator(bundle,
					wsdlConfig);
			reader.setFeature("javax.wsdl.verbose", false);
			def = reader.readWSDL(locator);

		}

		catch (Exception error) {
			throw new RuntimeException(error);
		}

//		find the port type named in the descriptor and process
//		all of the operations

		QName portTypeName = wsdlConfig.getWsdlPortType();
		PortType portType = def.getPortType(portTypeName);

		if (portType == null) // FIXME: extract message
			throw new RuntimeException("No port type of name " + portType);

		Collection ops = portType.getOperations();
		Iterator i = ops.iterator();

		Map opsByAction = new HashMap();

		while (i.hasNext())
		{
			Operation next = (Operation)i.next();
			String action = WsdlUtils.getAction(next);
			opsByAction.put(action, next);
		}

		return opsByAction;
	}


	/**
	 * factory method for creating Capability Descriptors
	 * @return a new OSGiCapabilityDescriptor
	 */
	protected CapabilityDescriptor createCapabilityDescriptor() {
		return new OSGiCapabilityDescriptor(bundle);
	}

	/**
	 * factory method for creating ResourceDefinitions
	 * @return a new OSGiResourceDefinition
	 */
    protected ResourceDefinition createResourceDefinition(){
    	return new OSGiResourceDefinition(bundle);
    }


}
