/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/


package org.apache.muse.core.platform.osgi.util;

import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;

import org.apache.muse.util.ReflectUtilHelper;
import org.apache.muse.util.osgi.CapabilityPair;
import org.apache.muse.util.osgi.ResourceManagementImplementation;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;

/**
 *
 * OSGiReflectUtilHelper is an implementation of the <code>ReflectUtilHelper</code> interface,
 * which allows platforms to extend the behavior of the ReflectUtils class.
 *
 * Additonally, this class acts as a repository for Resource and Capability implementations
 * that may be shared between bundles - simplifying the management of bundle dependencies when
 * using default capability and resource implementations provide by Muse.
 *
 * The repository is initialized by locating instance of ResourceManagementImplementation services
 * exported by bundles with contributions to make to the helper class. Most of the muse-*-impl or
 * muse-*-api bundle activators are contribution providers.
 *
 * @see ReflectUtilHelper
 * @see org.apache.muse.util.ReflectUtils
 * @see org.apache.muse.util.osgi.ResourceManagementImplementation
 * @author Joel Hawkins (joelh)
 *
 */

public class OSGiReflectUtilHelper implements ReflectUtilHelper, ServiceListener {

	private ThreadLocal localEnvironmentContext = new ThreadLocal();

	private static OSGiReflectUtilHelper _helper = new OSGiReflectUtilHelper();

	private BundleContext context;

	private HashMap contributorMap = new HashMap(); //Maps impl name to service
	private HashMap serviceMap = new HashMap(); //Maps reference to impl name

	//Map contributions to bundles
	private HashMap spec2BundleMap = new HashMap();
	private HashMap resourceImpl2BundleMap = new HashMap();
	private HashMap capabilityImpl2BundleMap = new HashMap();

	private static final String MANAGEMENT_IMPLEMENTATION_FILTER = "(objectclass=org.apache.muse.util.osgi.ResourceManagementImplementation)";

	private static class CapabilityEntry {

		public CapabilityEntry(String uri, String implName, Bundle bundle){
			this.uri = uri;
			this.implName = implName;
			this.bundle = bundle;
		}

		public String uri;
		public String implName;
		public Bundle bundle;
	}

	/**
	 * setter for the thread local bundle for the executing thread. This bundle will be
	 * used for reflection requests delagated to this helper by the executing thread.
	 * @param bundle
	 */
	public void setThreadLocalBundle(Bundle bundle) {
		localEnvironmentContext.set(bundle);
	}

	/**
	 * getter for the thread local bundle of the executing thread.
	 * @return the thread local Bundle for the executing thread.
	 */
	public Bundle getThreadLocalBundle() {
		return (Bundle)localEnvironmentContext.get();
	}

	/**
	 * access to the singleton instance of the OSGiReflectUtilHelper
	 * @return
	 */
	public static OSGiReflectUtilHelper getDefault() {
		return _helper;
	}

	/**
	 * initializes the helper class by locating all instances of <code>ResourceManagementImplementation</code>
	 * services and loading their contributions into the helper class.
	 * @param context
	 */
	public void initialize(BundleContext context){
		this.context = context;
		try {
			context.addServiceListener(this,MANAGEMENT_IMPLEMENTATION_FILTER);
			ServiceReference[] refs = context.getServiceReferences("org.apache.muse.util.osgi.ResourceManagementImplementation", MANAGEMENT_IMPLEMENTATION_FILTER);
			if(refs != null){
				for(int i=0;i<refs.length;i++){
					ResourceManagementImplementation impl = (ResourceManagementImplementation)context.getService(refs[i]);
					if(impl != null) loadContributions(impl);
				}
			}
		} catch (InvalidSyntaxException e) {
			//This should never happen as the filter is not dynamic
		}
	}

	private OSGiReflectUtilHelper() {
	}

	/**
	 * existence check for a class
	 * @param className the name of the class in question
	 * return <code>true</code> if the class exists in the executing thread's bundle context,
	 * otherwise <code>false</code>
	 */
	public boolean exists(String className) {
		Bundle bundle = (Bundle) localEnvironmentContext.get();
		if (bundle == null)
			return false;
		try {
			bundle.loadClass(className);
			return true;
		} catch (Throwable t) {
			try{
				OSGiReflectUtilHelper.class.getClassLoader().loadClass(className);
				return true;
			} catch(Throwable cnfe){
				return false;
			}
		}
	}

	private Class getClassFromContributions(String className){
		Bundle contributionBundle = null;
		contributionBundle = (Bundle) resourceImpl2BundleMap.get(className);
		if(contributionBundle == null){
			CapabilityEntry entry = (CapabilityEntry)capabilityImpl2BundleMap.get(className);
			if(entry != null) contributionBundle = entry.bundle;
		}
		if(contributionBundle == null) return null;
		try{
			return contributionBundle.loadClass(className);
		}catch(ClassNotFoundException cnfe){
			return null;
		}
		catch(Throwable t){
			return null;
		}
	}

	/**
	 * attempts to load a Class from the executing thread's bundle context.
	 * If the class is not found, an attempt is made to resolve the class based
	 * on contributions made by other bundes.
	 *
	 * @param className the name of the class in question
	 * @return the <code>Class</code> if the class exists in the executing thread's bundle context,
	 * otherwise <code>null</code>
	 */
	public Class getClass(String className) {
		Bundle bundle = (Bundle) localEnvironmentContext.get();
		if (bundle == null)
			return getClassFromContributions(className);
		try {
			return bundle.loadClass(className);
		} catch (Throwable t) {
			try{
				return OSGiReflectUtilHelper.class.getClassLoader().loadClass(className);
			} catch(Throwable cnfe){
				return getClassFromContributions(className);
			}
		}
	}

	/**
	 * attempts to resolve a resource from contributions made by other bundles. These resources
	 * are typically wsdl or schema files.
	 * @param name
	 * @return
	 */
	public InputStream getResource(String name){
		//Check against contributions -
		Bundle bundle = null;
		String trimmedName = name;
		synchronized(spec2BundleMap){
			bundle = (Bundle)spec2BundleMap.get(name);
			if(bundle == null && name.lastIndexOf("/") != -1){
				trimmedName = name.substring(name.lastIndexOf("/")+1);
				bundle = (Bundle)spec2BundleMap.get(trimmedName);
			}
		}
		if(bundle != null){
			try{
				URL url = bundle.getResource("/" + trimmedName);
				if(url != null) return url.openStream();
			}catch(Throwable t){
				//fall through
			}
		}
		return null;
	}

	private void loadContributions(ResourceManagementImplementation impl) {

		Bundle bundle = impl.getBundle();

		String implName = impl.getImplementationName();

		String[] resourceImplNames = impl.getResourceImplementationNames();
		String[] specNames = impl.getResourceSpecNames();
		CapabilityPair capabilities[] = impl.getResourceCapabilityMappings();
		int i;

		if(resourceImplNames != null){
			synchronized(resourceImpl2BundleMap){
				for(i=0;i<resourceImplNames.length;i++){
					resourceImpl2BundleMap.put(resourceImplNames[i], bundle);
				}
			}
		}
		if(specNames != null){
			synchronized(spec2BundleMap){
				for(i=0;i<specNames.length;i++){
					spec2BundleMap.put(specNames[i], bundle);
				}
			}
		}
		if(capabilities != null){
			synchronized(capabilityImpl2BundleMap){
				for(i=0;i<capabilities.length;i++){
					String uri = (String)capabilities[i].capabilityURI;
					String capImpl = (String)capabilities[i].capabilityImplementation;
					CapabilityEntry entry = new CapabilityEntry(uri,capImpl,bundle);
					capabilityImpl2BundleMap.put(capImpl, entry);
				}
			}
		}
	}


	/**
	 * ServiceListener method for dealing with ResourceManagementImplementation
	 * service lifecycle events.
	 */
	public synchronized void serviceChanged(ServiceEvent event) {
		ServiceReference ref = event.getServiceReference();
		ResourceManagementImplementation impl = (ResourceManagementImplementation)context.getService(ref);
		String name = impl.getImplementationName();
		switch (event.getType()) {
		case ServiceEvent.REGISTERED:
			if(contributorMap.get(name) == null){
				contributorMap.put(name, impl);
				serviceMap.put(ref, name);
				loadContributions(impl);
			}
			break;
		case ServiceEvent.MODIFIED:
			break;
		case ServiceEvent.UNREGISTERING:
			if (serviceMap.get(ref) != null) {
				contributorMap.remove(serviceMap.get(ref));
				serviceMap.remove(ref);
				//TODO clean up contributions
			}
			break;
		}

	}

}
