/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.tools.generator.projectizer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Logger;

import org.apache.muse.tools.generator.Wsdl2Java;
import org.apache.muse.util.FileUtils;
import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.addressing.WsaConstants;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * AbstractProjectizer contains methods that are useful to projectizers built so far.
 *
 * @author Andrew Eberbach (aeberbac)
 */
public abstract class AbstractProjectizer implements Projectizer {
	
	private static Messages _MESSAGES = MessagesFactory.get(AbstractProjectizer.class);
	
	private static Logger _logger = Logger.getLogger(Wsdl2Java.class.getPackage().getName()); 
	
	protected static final String DEFAULT_WSDL_NAME_SUFFIX = ".wsdl";

	protected static final String PLACE_HOLDER = "XXX";

	protected boolean _overwrite = false;
	
	private int _resourceCounter = 1; 	
	
	/**
	 * Serializes an XML node and writes it out to a file provided that
	 * the file doesn't already exist, and if it does then the _overwrite 
	 * variable must be set to true.
	 * 
	 * @param node 
	 * 			The DOM node to write out
	 * @param destination 
	 * 			The file where this should be written
	 * 
	 * @throws Exception
	 * 			If anything goes wrong
	 */
	protected void writeToFileCheck(Node node, File destination) throws Exception {
		if (!destination.exists() || _overwrite) {
			File parent = destination.getAbsoluteFile().getParentFile();
			if(!parent.exists()) {
				if(!parent.mkdirs()) {
					Object[] filler = { parent } ;
					throw new Exception(_MESSAGES.get("CouldNotMakeDir",filler));
				}
			}
			
			FileWriter fileWriter = new FileWriter(destination);
			fileWriter.write(XmlUtils.toString(node));
			fileWriter.close();
		} else {
			_logger.info(_MESSAGES.get("NotOverwriting", new String[] {destination.getPath()}));
		}
	}
	
	/**
	 * Writes a string to a file provided that
	 * the file doesn't already exist, and if it does then the _overwrite 
	 * variable must be set to true.
	 * 
	 * @param string 
	 * 			The string to write out
	 * @param destination 
	 * 			The file where this should be written
	 * 
	 * @throws Exception
	 * 			If anything goes wrong
	 */
	protected void writeToFileCheck(String string, File destination) throws Exception {
		if (!destination.exists() || _overwrite) {
			File parent = destination.getAbsoluteFile().getParentFile();
			if(!parent.exists()) {
				if(!parent.mkdirs()) {
					Object[] filler = { parent } ;
					throw new Exception(_MESSAGES.get("CouldNotMakeDir",filler));
				}
			}
			
			FileWriter fileWriter = new FileWriter(destination);
			fileWriter.write(string);
			fileWriter.close();
		} else {
			_logger.info(_MESSAGES.get("NotOverwriting", new String[] {destination.getPath()}));
		}
	}
	

	/**
	 * Writes an <code>InputStream</code> to a file provided that
	 * the file doesn't already exist, and if it does then the _overwrite 
	 * variable must be set to true.
	 * 
	 * @param inputStream 
	 * 			The <code>InputStream</code> to write out
	 * @param destination 
	 * 			The file where this should be written
	 * 
	 * @throws Exception
	 * 			If anything goes wrong
	 */
	protected void copyStreamCheck(InputStream inputStream, File destination) throws Exception {
		if(!destination.exists() || _overwrite) {
			File parent = destination.getAbsoluteFile().getParentFile();
			if(!parent.exists()) {
				if(!parent.mkdirs()) {
					Object[] filler = { parent } ;
					throw new Exception(_MESSAGES.get("CouldNotMakeDir",filler));
				}
			}
			
			FileUtils.copyFile(inputStream,destination);		
		} else {
			_logger.warning(_MESSAGES.get("NotOverwriting", new String[] {destination.getPath()}));
		}
	}
	
	/**
	 * Read an <code>InputStream</code> replacing all occurences of PLACE_HOLDER
	 * with the values provided in the filler.
	 * 
	 * @param inputStream 
	 * 				The source input
	 * @param filler
	 * 				The filler for the string template 
	 * 
	 * @return
	 * 				A string with as many PLACE_HOLDERs replaced as possible
	 * 
	 * @throws Exception
	 * 				If anything goes wrong
	 */
	protected String loadString(InputStream inputStream, Object[] filler) throws Exception {
		StringBuffer result = new StringBuffer();
		BufferedReader br = new BufferedReader(new InputStreamReader(inputStream));
		
		String line = null;
		int i=0;
		
		while( (line = br.readLine()) != null) {
			if(line.indexOf(PLACE_HOLDER) > 0) {
				if(i < filler.length) {
					line = line.replaceFirst(PLACE_HOLDER, filler[i++].toString());
				}
			}
			result.append(line + "\n");
		}
		
		return result.toString();
	}
	
	/**
	 * Copies a tempalte folder to another folder. The idea here is 
	 * since we want to copy an entire folder it, it will be created
	 * <b>under</b> the destination folder. So this method tries to
	 * copy the folder to the parent and then renames it to the name of the
	 * destination. 
	 * 
	 * For example. Say you have a template folder called "c:\src\foo" and you
	 * want to copy it to a destination called "c:\dest\bar". You would copy "c:\src\foo"
	 * to the parent of the destination, ie "c:\dest" creating "c:\dest\foo" and
	 * then you would rename the folder to "c:\dest\bar".
	 * 
	 * If the target folder exists we log a warning and stop because it's
	 * not entirely obvious what users would expect even if _overwrite is true.
	 * 
	 * @param source
	 * 			The source folder
	 * @param destination
	 * 			The destination, ie what the folder will be called 
	 * 
	 * @return
	 * 			The destination
	 * 
	 * @throws IOException
	 * 			If anything goes wrong
	 */
	protected File copyTemplate(File source, File destination) throws IOException {
		if(destination.exists()) {
			Object[] filler = { destination.getAbsolutePath() };
			_logger.warning(_MESSAGES.get("DirectoryExists", filler));
			return destination;
		}
		
		File parent = destination.getAbsoluteFile().getParentFile();
		
		FileUtils.copyDirectory(source, parent);
        
        File templateDir = new File(parent, source.getName());                    
        
        templateDir.renameTo(destination);
        
        return destination;
	}
	
	/**
	 * Create a router entry for the given service in the given directory. This
	 * will generate a router entry using a simple integer counter for the file name
	 * and it will contain an empty ReferenceParameters element.
	 * 
	 * @param routerEntriesDir
	 * 			The directory where the router entries should be created
	 * @param serviceName
	 * 			The name of the service
	 *  
	 * @throws Exception
	 * 			If anything goes wrong
	 */
	protected void createRouterEntries(File routerEntriesDir, String serviceName) throws Exception{
		Document routerEntryDocument = XmlUtils.createDocument();
		Element parametersElement = XmlUtils.createElement(routerEntryDocument, WsaConstants.PARAMETERS_QNAME);
		routerEntryDocument.appendChild(parametersElement);
		File serviceDir = new File(routerEntriesDir, serviceName);
		File routerEntryFile = new File(serviceDir, getResourceFileName());
		writeToFileCheck(routerEntryDocument, routerEntryFile);
	}
	
	/**
	 * Get the name of the next router entry file. This is done
	 * by using an integer counter and creating a name out of it.
	 * 
	 * For example:
	 *  	resource-instance-1.xml
	 *  	resource-instance-2.xml
	 *  
	 *  and so on.
	 * 
	 * @return	The next available name
	 */
	private String getResourceFileName() {
		return Axis2ProjectizerConstants.RESOURCE_FILE.replaceFirst(PLACE_HOLDER, String.valueOf(_resourceCounter++));
	}

	/**
	 * Given a list of Maps (mapping file names to file contents) write the files
	 * to the target directory.
	 * 
	 * @param javaSourceDir
	 * 				The target directory
	 * 
	 * @param filesMaps
	 * 				A list of Maps which map file names to file content
	 * 			
	 * @throws Exception
	 * 				If anything goes wrong
	 */
	protected void createJavaSources(File javaSourceDir, Map[] filesMaps) throws Exception {
		for(int i = 0; i < filesMaps.length; i++) {
			Map files = filesMaps[i];
			
			for (Iterator j = files.keySet().iterator(); j.hasNext();) {				
				String file = (String) j.next();
	
				File javaFile = new File(javaSourceDir, file);			
	
				writeToFileCheck((String) files.get(file), javaFile);
			}
		}
	}
	
	/**
	 * Utility method for copying jars from the Muse distribution. The
	 * module names serve as names of directories (ie. core, tools, and so on). The
	 * baseModulesDir is where to start from (ie. c:\muse-2.0\modules). All of the jars 
	 * in each modulename directory are copied to the destDir.
	 * 
	 * @param moduleNames
	 * 			The modules to copy
	 * 
	 * @param baseModulesDir
	 * 			The location where the modules reside
	 * 
	 * @param destDir
	 * 			The destination directory
	 * 
	 * @throws IOException
	 * 			If anything goes wrong
	 */
	protected void copyJars(String[] moduleNames, File baseModulesDir, File destDir) throws IOException {
		if(!baseModulesDir.exists()) {
			throw new FileNotFoundException();
		}
		
		if(!destDir.exists()) {
			if(!destDir.mkdirs()) {
				throw new RuntimeException();
			}
		}
	
		File curDir = null;
		File curFile = null;
		
		for(int i=0; i < moduleNames.length; i++) {
			curDir = new File(baseModulesDir, moduleNames[i]);
			if(!curDir.exists()) {
				throw new FileNotFoundException();
			}
			
			String[] files = curDir.list();
			
			for(int j=0; j < files.length; j++) {
				if(files[j].toLowerCase().endsWith(".jar")) {
					curFile = new File(curDir, files[j]);
					FileUtils.copy(curFile, destDir);
				}
			}
		}
	}
}
