/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.dm.muws.events.impl;

import java.text.ParseException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.dm.muws.events.Situation;
import org.apache.muse.ws.dm.muws.events.SubstitutableMessage;
import org.apache.muse.ws.dm.muws.events.WefConstants;

/**
 * 
 * @author Dan Jemiolo (danj)
 *  
 */

public class SimpleSituation implements Situation
{
    //
    // used to lookup all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(SimpleSituation.class);
    
    private QName[] _categoryType = null;

    private String _message = null;
    
    private short _priority = -1;

    private short _severity = -1;

    private Date _situationTime = new Date();

    private SubstitutableMessage _subMessage = null;

    private String _successDisposition = null;
    
    public SimpleSituation()
    {
        //
        // no action - allow users to create situations from scratch
        //
    }
    
    public SimpleSituation(Element xml)
    {
        //
        // NOTE: we can use public get/set methods in the constructor 
        //       because they are final
        //
        setMessage(XmlUtils.getElementText(xml, WefConstants.MESSAGE_QNAME));
        
        String disposition = XmlUtils.getElementText(xml, WefConstants.SUCCESS_DISPOSITION_QNAME);
        
        if (disposition != null)
            setSuccessDisposition(disposition);
        
        Element priorityXML = XmlUtils.getElement(xml, WefConstants.PRIORITY_QNAME);
        
        if (priorityXML != null)
            _priority = XmlUtils.getShort(priorityXML).shortValue();
        
        Element severityXML = XmlUtils.getElement(xml, WefConstants.SEVERITY_QNAME);
        
        if (severityXML != null)
            _severity = XmlUtils.getShort(severityXML).shortValue();
        
        Element timeXML = XmlUtils.getElement(xml, WefConstants.SITUATION_TIME_QNAME);
        
        try
        {
            if (timeXML != null)
                _situationTime = XmlUtils.getDate(timeXML);
        }
        
        catch (ParseException error)
        {
            throw new RuntimeException(error.getMessage(), error);
        }
        
        Element categoryXML = XmlUtils.getElement(xml, WefConstants.SITUATION_CATEGORY_QNAME);
        Element categoryTypeXML = XmlUtils.getFirstElement(categoryXML);
        
        List hierarchy = new LinkedList();
        
        while (categoryTypeXML != null)
        {
            QName categoryName = XmlUtils.getElementQName(categoryTypeXML);
            hierarchy.add(categoryName);
            categoryTypeXML = XmlUtils.getElement(categoryTypeXML, WefConstants.SITUATION_CATEGORY_QNAME);
        }
        
        _categoryType = new QName[hierarchy.size()];
        _categoryType = (QName[])hierarchy.toArray(_categoryType);
    }

    public final QName[] getCategoryType()
    {
        return _categoryType;
    }

    public final String getMessage()
    {
        return _message;
    }

    public final short getPriority()
    {
        return _priority;
    }
    
    public final short getSeverity()
    {
        return _severity;
    }

    public final Date getSituationTime()
    {
        return _situationTime;
    }

    public final SubstitutableMessage getSubstitutableMessage()
    {
        return _subMessage;
    }

    public final String getSuccessDisposition()
    {
        return _successDisposition;
    }

    public final void setCategoryType(QName categoryType)
    {
        setCategoryType(new QName[]{ categoryType });
    }
    
    public final void setCategoryType(QName[] categoryType)
    {
        if (categoryType == null || categoryType.length == 0)
            throw new IllegalArgumentException(_MESSAGES.get("NoCategoryType"));
        
        _categoryType = categoryType;
    }

    public final void setMessage(String message)
    {
        _message = message;
    }

    public final void setPriority(short priority)
    {
        if (priority < LOWEST_PRIORITY || priority > HIGHEST_PRIORITY)
        {
            Object[] filler = { new Short(priority), new Short(LOWEST_PRIORITY), new Short(HIGHEST_PRIORITY) };
            throw new IllegalArgumentException(_MESSAGES.get("PriorityOutOfBounds", filler));
        }
        
        _priority = priority;
    }

    public final void setSeverity(short severity)
    {
        if (severity < UNKNOWN_SEVERITY || severity > FATAL_SEVERITY)
        {
            Object[] filler = { new Short(severity), new Short(UNKNOWN_SEVERITY), new Short(FATAL_SEVERITY) };
            throw new IllegalArgumentException(_MESSAGES.get("SeverityOutOfBounds", filler));
        }
        
        _severity = severity;
    }

    public final void setSituationTime(Date situationTime)
    {
        _situationTime = situationTime;
    }
    
    public final void setSubstitutableMessage(SubstitutableMessage message)
    {
        _subMessage = message;
    }

    public final void setSuccessDisposition(String disposition)
    {
        if (!UNSUCCESSFUL.equals(disposition) && !SUCCESSFUL.equals(disposition))
            throw new IllegalArgumentException(_MESSAGES.get("InvalidDisposition", new Object[]{ disposition }));
        
        _successDisposition = disposition;
    }
    
    public String toString()
    {
        return XmlUtils.toString(toXML(), false);
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document factory)
    {
        Element situation = XmlUtils.createElement(factory, WefConstants.SITUATION_QNAME);

        QName[] type = getCategoryType();

        if (type == null || type.length == 0)
            throw new IllegalArgumentException(_MESSAGES.get("NoCategoryType"));
        
        Element categoryType = XmlUtils.createElement(factory, WefConstants.SITUATION_CATEGORY_QNAME);
        situation.appendChild(categoryType);
        
        Element parent = categoryType;
        
        for (int n = 0; n < type.length; ++n)
        {
            Element nextPart = XmlUtils.createElement(factory, type[n]);
            parent.appendChild(nextPart);
            parent = nextPart;
        }
        
        String disposition = getSuccessDisposition();
        
        if (disposition != null)
            XmlUtils.setElement(situation, WefConstants.SUCCESS_DISPOSITION_QNAME, disposition);

        Date time = getSituationTime();
        
        if (time != null)
            XmlUtils.setElement(situation, WefConstants.SITUATION_TIME_QNAME, time);

        short priority = getPriority();
        
        if (priority >= 0)
            XmlUtils.setElement(situation, WefConstants.PRIORITY_QNAME, new Short(priority));
        
        short severity = getSeverity();
        
        if (severity >= 0)
            XmlUtils.setElement(situation, WefConstants.SEVERITY_QNAME, new Short(severity));
        
        String message = getMessage();

        if (message != null)
            XmlUtils.setElement(situation, WefConstants.MESSAGE_QNAME, message);
        
        SubstitutableMessage subMessage = getSubstitutableMessage();
        
        if (subMessage != null)
            XmlUtils.setElement(situation, WefConstants.SUBSTITUTABLE_MSG_QNAME, subMessage);
        
        return situation;
    }
}
