/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification.topics;

import java.util.Collection;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.resource.basefaults.BaseFault;

/**
 * 
 * Topic is an interface that represents the wsnt:Topic data structure described 
 * in WS-Notification v1.3 and WS-Topics v1.3. Instances of this type are 
 * collected by WS-N NotificationProducer implementations to implement topic 
 * filters on subscriptions.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public interface Topic extends XmlSerializable
{
    void addMessageType(QName messageType);

    /**
     * 
     * Adds a child topic to the topic.
     *
     * @param childTopic
     *
     */
    void addTopic(Topic childTopic) 
        throws BaseFault;
    
    /**
     *
     * @return A fully-qualified name for the topic.
     *
     */
    QName getConcretePath();

    /**
     *
     * @return The last message published to this topic.
     *
     */
    NotificationMessage getCurrentMessage();
    
    String getMessagePattern();
    
    String getMessagePatternDialect();

    Set getMessageTypes();

    /**
     *
     * @return The local name of the topic.
     *
     */
    String getName();
    
    Topic getParentTopic();

    /**
     * 
     * @param topicName
     * 
     * @return The child topic with the given name, or null if no such topic exists.
     *
     */
    Topic getTopic(String topicName);

    /**
     * 
     * @return The TopicNamespace that contains this Topic.
     *
     */
    TopicNamespace getTopicNamespace();

    /**
     * 
     * @return All of the child topics under this topic.
     *
     */
    Collection getTopics();

    /**
     * 
     * @param topicName
     * 
     * @return True if this topic has a child topic with the given name.
     *
     */
    boolean hasTopic(String topicName);

    /**
     * 
     * @return True if the topic does not allow addition of child topics.
     *
     */
    boolean isFinal();

    /**
     *
     * @return True if the topic has no parent topic.
     *
     */
    boolean isRootTopic();

    void removeAllTopics();

    void removeMessageType(QName messageType);

    void removeTopic(String topicName);
    
    /**
     * 
     * @param message
     *        The last message published to this topic.
     *
     */
    void setCurrentMessage(NotificationMessage message);

    void setFinal(boolean isFinal);

    void setMessagePattern(String messagePattern);

    void setParentTopic(Topic parentTopic) 
        throws BaseFault;
}
