/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.notification.impl;

import org.apache.muse.ws.addressing.EndpointReference;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.ws.notification.Filter;
import org.apache.muse.ws.notification.NotificationConsumer;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.notification.NotificationMessageListener;
import org.apache.muse.ws.notification.PullPoint;
import org.apache.muse.ws.notification.PullPointDataStore;
import org.apache.muse.ws.notification.WsnConstants;
import org.apache.muse.ws.notification.faults.UnableToGetMessagesFault;
import org.apache.muse.ws.resource.WsResource;
import org.apache.muse.ws.resource.impl.AbstractWsResourceCapability;

/**
 * 
 * SimplePullPoint is Muse's default implementation of the WS-Notification 
 * PullPoint port type and the data store API that augments it. Users that 
 * want to change the way that messages are stored should override the 
 * createDataStore() method to return an alternate implementation of the 
 * {@linkplain PullPointDataStore PullPointDataStore API}. The default 
 * implementation is {@linkplain SimplePullPointDataStore SimplePullPointDataStore}.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class SimplePullPoint 
    extends AbstractWsResourceCapability implements PullPoint, NotificationMessageListener
{
    //
    // the place where we keep messages until a client asks for them
    //
    private PullPointDataStore _dataStore = null;
    
    //
    // the subscription that is sending messages to our pullpoint
    //
    private WsResource _subscription = null;
    
    /**
     * 
     * @return True if the message has a subscription EPR that matches our 
     *         subscription's EPR
     *         
     */
    public boolean accepts(NotificationMessage message)
    {
        EndpointReference subEPR = message.getSubscriptionReference();
        return subEPR != null && getSubscription().equals(subEPR);
    }
    
    /**
     * 
     * Users can override this method to provide an alternate implementation 
     * of the PullPointDataStore API.
     *
     * @return An instance of SimplePullPointDataStore
     *
     */
    protected PullPointDataStore createDataStore()
    {
        return new SimplePullPointDataStore();
    }

    public PullPointDataStore getDataStore()
    {
        return _dataStore;
    }
    
    /**
     * 
     * @return An instance of PublishAllMessagesFilter. This allows the pullpoint 
     *         to add all messages to its data store.
     * 
     */
    public Filter getFilter()
    {
        return PublishAllMessagesFilter.getInstance();
    }

    public NotificationMessage[] getMessages(int maxNumber) 
        throws UnableToGetMessagesFault
    {
        return getDataStore().getMessages(maxNumber);
    }
    
    public WsResource getSubscription()
    {
        return _subscription;
    }

    public void initialize() 
        throws SoapFault
    {
        super.initialize();
        
        _dataStore = createDataStore();
        _dataStore.initialize();
        
        //
        // register for messages so we can add them to the data store
        //
        NotificationConsumer wsn = (NotificationConsumer)getResource().getCapability(WsnConstants.CONSUMER_URI);
        wsn.addMessageListener(this);
    }
    
    public void process(NotificationMessage message) 
        throws SoapFault
    {
        getDataStore().addMessage(message);
    }

    public void setSubscription(WsResource subscription)
    {
        _subscription = subscription;
    }

    public void shutdown() 
        throws SoapFault
    {
        getSubscription().shutdown();
        getDataStore().shutdown();
        super.shutdown();
    }
}
