/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
 
package org.apache.muse.ws.notification.properties;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.xml.XmlSerializable;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.notification.NotificationMessage;
import org.apache.muse.ws.resource.properties.WsrpConstants;

/**
 * 
 * ResourcePropertyChangeEvent is a simple Java bean that represents the WSRP 
 * property change notification event. Users can instantiate this class and 
 * provide it to NotificationProducer.publish() in order to report property 
 * changes that aren't caught by the Muse WSRP implementation.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class ResourcePropertyChangeEvent implements XmlSerializable
{
    private Element[] _newValues = null;
    
    private Element[] _oldValues = null;
    
    public ResourcePropertyChangeEvent(Element[] oldValues, Element[] newValues)
    {
        _oldValues = oldValues;
        _newValues = newValues;
    }
    
    /**
     * 
     * Parses the message content to find the WSRP change notification data.
     * 
     * @param message
     *        A notification message containing a WSRP change event.
     *
     */
    public ResourcePropertyChangeEvent(NotificationMessage message)
    {
        Element wsrpXML = message.getMessageContent(WsrpConstants.NOTIFICATION_QNAME);
        Element oldValuesXML = XmlUtils.getElement(wsrpXML, WsrpConstants.OLD_VALUES_QNAME);
        Element newValuesXML = XmlUtils.getElement(wsrpXML, WsrpConstants.NEW_VALUES_QNAME);
        
        _oldValues = XmlUtils.getAllElements(oldValuesXML);
        _newValues = XmlUtils.getAllElements(newValuesXML);
    }
    
    public Element[] getNewValues()
    {
        return _newValues;
    }
    
    public Element[] getOldValues()
    {
        return _oldValues;
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document doc)
    {
        Element root = XmlUtils.createElement(doc, WsrpConstants.NOTIFICATION_QNAME);
        
        Element[] oldValues = getOldValues();
        Element[] newValues = getNewValues();
        
        Element oldValueXML = XmlUtils.createElement(doc, WsrpConstants.OLD_VALUES_QNAME);
        Element newValueXML = XmlUtils.createElement(doc, WsrpConstants.NEW_VALUES_QNAME);
        root.appendChild(oldValueXML);
        root.appendChild(newValueXML);
        
        for (int n = 0; n < oldValues.length; ++n)
        {
            oldValues[n] = (Element)doc.importNode(oldValues[n], true);
            oldValueXML.appendChild(oldValues[n]);
        }
        
        for (int n = 0; n < newValues.length; ++n)
        {
            newValues[n] = (Element)doc.importNode(newValues[n], true);
            newValueXML.appendChild(newValues[n]);
        }
        
        return root;
    }
}
