/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.metadata;

import org.apache.muse.ws.resource.basefaults.BaseFault;

/**
 * 
 * ResourcePropertiesMetadataValidation is a set of methods required of 
 * WSRP implementations that provide metadata validation via WS-RMD 1.0.
 * 
 * @author Dan Jemiolo (danj)
 *
 */

public interface ResourcePropertiesMetadataValidation
{
    /**
     * 
     * Takes the RMD that has been provided through setMetadata() and creates 
     * components that will enforce it. This method only needs to be called 
     * once after setMetadata() is called. After this method is called, any 
     * attempt to modify the WS-RP container in a way that is incompatible 
     * with the RMD will fail. 
     *
     */
    void applyMetadata()
        throws BaseFault;
    
    /**
     * 
     * @return The RMD for this WS-RP document.
     *
     */
    MetadataDescriptor getMetadata();
    
    /**
     * 
     * Provides the given WS-RMD MetadataDescriptor to this WS-RP document. 
     * This metadata will be used for all subsequent validations, including 
     * calls to validateMetadata() and SetResourceProperties-like operations.
     * <br><br>
     * Note that this method does <b>not</b> perform validation and it does 
     * <b>not</b> create listeners to do so in the future. To do these things, 
     * use validateMetadata and applyMetadata, respectively.
     *
     * @param metadata
     *        The WS-RMD metadata for this WS-RP document.
     * 
     * @see #applyMetadata()
     * @see #validateMetadata()
     *
     */
    void setMetadata(MetadataDescriptor metadata);
    
    /**
     * 
     * Iterates through every property in the WS-RMD metadata and verifies 
     * that the current WS-RP document is valid. This method should be used 
     * sparingly, as it will validate every property; for cases such as 
     * SetResourceProperties, the implementation of the operations should 
     * validate the individual property being targeted, not the whole document.
     *
     * @throws BaseFault
     *         <ul>
     *         <li>If there is a property in the WS-RMD document that is not 
     *         in the schema.</li>
     *         <li>If the WS-RMD defines a set of valid values for a property 
     *         and one of the current values is not included in that set.</li>
     *         </ul>
     *
     */
    void validateMetadata()
        throws BaseFault;
}
