/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.properties.listeners;

import javax.xml.namespace.QName;

import org.w3c.dom.Element;

import org.apache.muse.ws.resource.basefaults.BaseFault;

/**
 *
 * PropertyChangeApprover describes a type of property listener that is 
 * invoked when a property change is about to occur (through the WS-RP 
 * SetResourceProperties operation). Approvers can prevent the change by 
 * throwing an exception from the validateChange() method. Before writing 
 * your own approver, make sure that Muse's WS-RMD support doesn't already 
 * do what you are trying to do.
 *
 * @author Dan Jemiolo (danj)
 * 
 * @see org.apache.muse.ws.resource.metadata.MetadataDescriptor
 *
 */

public interface PropertyChangeApprover
{
    /**
     * 
     * @return The QName of the property being monitored.
     *
     */
    QName getPropertyName();
    
    /**
     * 
     * @param securityToken
     *        The security token provided by the caller who wishes to make 
     *        the property change.
     * 
     * @return True if the token matches one the approver knows to be valid.
     * 
     * @see #setSecurityToken(Object)
     *
     */
    boolean isSecure(Object securityToken);
    
    /**
     * 
     * Sets the security token that will be used in all security checks 
     * via isSecure(Object).
     *
     * @param securityToken
     * 
     * @see #isSecure(Object)
     *
     */
    void setSecurityToken(Object securityToken);
    
    /**
     * 
     * Analyzes the property change being made and throws an exception if 
     * the change is invalid.
     *
     * @param oldValue
     *        The current (about to be "old") value of the resource property.
     *        This should be null if the change is a WS-RP Insert.
     * 
     * @param newValue
     *        The potential new value of the resource property. This should 
     *        be null if the change is a WS-RP Delete.
     * 
     * @param token
     *        The security token provided by the caller who wishes to make 
     *        the property change.
     * 
     * @throws BaseFault
     *         <ul>
     *         <li>If the modification is not allowed for any reason.</li>
     *         </ul>
     *
     */
    void validateChange(Element oldValue, Element newValue, Object token)
        throws BaseFault;
}
