/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.properties.listeners;

import java.util.Iterator;

import javax.xml.namespace.QName;

/**
 *
 * ResourcePropertyListeners defines operations for WS-RP implementations 
 * that allow users to monitor activity on the document and subscribe to 
 * notifications about that activity. Because the concepts of validation 
 * and notifications are often used together, this interface also defines 
 * a set of operations for applying a schema and metadata restrictions to the 
 * WS-RP document.
 *
 * @author Dan Jemiolo (danj)
 * 
 * @see org.apache.muse.ws.resource.properties.ResourcePropertyCollection
 * 
 */

public interface ResourcePropertyListeners
{
    /**
     * 
     * Associates the given approver with the appropriate property. All 
     * approvers will be notified of pending changes to their property; if 
     * one of the approvers rejects the change (by throwing an exception), 
     * the change will not be completed. Each approver will continue to 
     * exist until it is removed with removeChangeApprover(PropertyChangeApprover). 
     *
     * @param approver
     *        The approver that will be notified of all pending changes.
     * 
     * @see PropertyChangeApprover#getPropertyName()
     * @see #removeChangeApprover(PropertyChangeApprover)
     *
     */
    void addChangeApprover(PropertyChangeApprover approver);
    
    /**
     * 
     * Associates the given listener with the appropriate property. The 
     * listeners will be notified of completed changes to their property. 
     * Each listener will continue to exist until it is removed with 
     * removeChangeListener(PropertyChangeListener).
     *
     * @param listener
     *        The listener that will be notified of all completed changes.
     * 
     * @see PropertyChangeListener#getPropertyName()
     * @see #removeChangeListener(PropertyChangeListener)
     *
     */
    void addChangeListener(PropertyChangeListener listener);

    /**
     * 
     * Associates the given listener with the appropriate property. The 
     * listeners will be notified of all pending read requests for a 
     * property, giving them the opportunity to reject the read or to modify 
     * the value of the property before it is returned to the caller. Each 
     * listener will continue to exist until it is removed with 
     * removeReadListener(PropertyReadListener).
     *
     * @param listener
     *        The listener that will be notified of all pending read requests.
     * 
     * @see PropertyReadListener#getPropertyName()
     * @see #removeReadListener(PropertyReadListener)
     *
     */
    void addReadListener(PropertyReadListener listener);
    
    /**
     * 
     * @return An iterator for the collection of PropertyChangeApprovers. 
     *         The approvers are not guaranteed to be in a particular order.
     *
     */
    Iterator getChangeApprovers(QName property);
    
    /**
     * 
     * @return An iterator for the collection of PropertyChangeListeners. 
     *         The listeners are not guaranteed to be in a particular order.
     *
     */
    Iterator getChangeListeners(QName property);
    
    /**
     * 
     * @return An iterator for the collection of PropertyReadListeners. 
     *         The listeners are not guaranteed to be in a particular order.
     *
     */
    Iterator getReadListeners(QName property);
    
    /**
     * 
     * Removes the assocation between the given approver and its property.
     * The approver will no longer be notified of pending changes.
     *
     * @param approver
     *
     */
    void removeChangeApprover(PropertyChangeApprover approver);
    
    /**
     * 
     * Removes the assocation between the given listener and its property.
     * The approver will no longer be notified of completed changes.
     *
     * @param listener
     *
     */
    void removeChangeListener(PropertyChangeListener listener);

    /**
     * 
     * Removes the assocation between the given listener and its property.
     * The approver will no longer be notified of pending read requests.
     *
     * @param listener
     *
     */
    void removeReadListener(PropertyReadListener listener);
}
