/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.properties.schema;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.apache.muse.ws.resource.WsResourceCapability;

/**
 *
 * ResourcePropertiesSchema represents the portion of an XML schema that 
 * defines a WS-RP document. Implementations of this interface provide a 
 * lookup service for property schema information.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public interface ResourcePropertiesSchema
{
    /**
     * 
     * The unbounded value can be applied to the property's minimum or 
     * maximum setting. An unbounded minimum means that a property can 
     * have zero instances; an unbounded maximum means that a property 
     * can have unlimited instances.
     * 
     */
    int UNBOUNDED = Integer.MAX_VALUE;
    
    /**
     * 
     * @param property
     * 
     * @return The capability that defines the resource property, or null if 
     *         the property does not exist.
     *
     */
    WsResourceCapability getCapability(QName property);
    
    /**
     * 
     * @return The name of the schema element representing the WSRP document.
     *
     */
    QName getElementName();
    
    /**
     * 
     * @param property
     *        
     * @return The maximum number of occurrences for the given property, or 
     *         UNBOUNDED if the property's "maxOccurs" is set to "unbounded".
     *
     */
    int getMaxOccurs(QName property);
    
    /**
     * 
     *
     * @param property
     * 
     * @return The minimum number of occurrences for the given property.
     *
     */
    int getMinOccurs(QName property);
    
    /**
     *
     * @return The collection of property names defined in this schema.
     *
     */
    Collection getPropertyNames();
        
    /**
     *
     * @param property
     * 
     * @return The qualified name of the property's type (set with the 
     *         "type" attribute). If no type is specified, the method 
     *         returns XsdUtils.ANY_TYPE_QNAME.
     *
     */
    QName getPropertyTypeName(QName property);
    
    boolean hasCapability(QName property);
    
    /**
     *
     * @param property
     * 
     * @return True is the schema has a property definition with the given name.
     *
     */
    boolean hasProperty(QName property);
    
    /**
     * 
     * @param property
     * 
     * @return True if getMaxOccurs(QName) returns UNBOUNDED.
     *
     */
    boolean isMaxUnbounded(QName property);
    
    /**
     * 
     * @param property
     * 
     * @return True if instances of the given property can be set to null, 
     *         or empty elements (set with the "nillable" attribute).
     *
     */
    boolean isNillable(QName property);
    
    /**
     * 
     * Maps the property to the capability so that users can later determine 
     * which capability should be used for servicing read/write requests 
     * against the property.
     * 
     * @param property
     * @param capability
     *
     */
    void setCapability(QName property, WsResourceCapability capability);
    
    /**
     * 
     * @param wsrpName
     *        The name of the schema element representing the WSRP document.
     *
     */
    void setElementName(QName wsrpName);
}
