/*=============================================================================*
 *  Copyright 2006 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/

package org.apache.muse.ws.resource.metadata.impl;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.ws.resource.metadata.MetadataDescriptor;
import org.apache.muse.ws.resource.metadata.WsrmdConstants;
import org.apache.muse.ws.resource.properties.impl.WsrpUtils;
import org.apache.muse.ws.addressing.soap.SoapFault;
import org.apache.muse.util.xml.XmlUtils;

/**
 * 
 * SimpleMetadataDescriptor is Muse's default implementation of the RMD
 * parsing and evaluation stage. It stores a simple lookup table of property
 * names and metadata.
 * 
 * @author Dan Jemiolo (danj)
 *  
 */

public class SimpleMetadataDescriptor implements MetadataDescriptor
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(SimpleMetadataDescriptor.class);
    
    //
    // The (unique) name of the descriptor - this maps to the name attribute
    // in the XML definition
    //
    private String _name = null;
    
    //
    // All of the wsrmd:Property definitions, according to paths (a QName)
    //
    private Map _propertiesByQName = new HashMap();

    private QName _wsdlInterface = null;

    private String _wsdlLocation = null;
    
    /**
     * 
     * Creates a new RMD metadata descriptor from the given XML definition.
     * 
     * @param xml
     *        The XML representation of the descriptor.
     *  
     */
    public SimpleMetadataDescriptor(Element xml)
    {
        if (xml == null)
            throw new NullPointerException(_MESSAGES.get("NullDescriptorElement"));

        //
        // the (unique) name - this is required
        //
        _name = xml.getAttribute(WsrmdConstants.NAME);

        if (_name == null || _name.length() == 0)
            throw new RuntimeException(_MESSAGES.get("NoDescriptorName"));
        
        String interfaceString = xml.getAttribute(WsrmdConstants.INTERFACE);

        if (interfaceString == null || interfaceString.length() == 0)
            throw new RuntimeException(_MESSAGES.get("NoDescriptorInterface"));
        
        _wsdlInterface = XmlUtils.parseQName(interfaceString, xml);
        
        String uriList = xml.getAttribute(WsrmdConstants.WSDL_LOCATION);

        if (uriList == null || uriList.length() == 0)
            throw new RuntimeException(_MESSAGES.get("NoDescriptorWSDLLocation"));
        
        //
        // the list is a two-URI set of { WSDL targetNS, WSDL location }
        //
        String[] values = uriList.split(" ");
        
        if (values.length != 2)
        {
            Object[] filler = { uriList };
            throw new RuntimeException(_MESSAGES.get("InvalidWSDLLocation", filler));
        }
            
        _wsdlLocation = values[1];  // ignore values[0] - it's just the targetNS

        //
        // parse all properties (there may be zero)
        //
        Element[] properties = XmlUtils.getElements(xml, WsrmdConstants.PROPERTY_QNAME);

        for (int n = 0; n < properties.length; ++n)
        {
            PropertyMetadata metadata = new PropertyMetadata(properties[n]);
            _propertiesByQName.put(metadata.getPropertyName(), metadata);
        }
    }

    public boolean canDelete(QName propertyQName)
    {
        //
        // deletion requires mutability (AND isReadOnlyExternal() == false)
        //
        PropertyMetadata property = getProperty(propertyQName);
        return property.getMutability() == PropertyMetadata.MUTABLE;
    }

    public boolean canInsert(QName propertyQName)
    {
        //
        // insert requires write permission and partial mutability (AND
        // isReadOnlyExternal() == false)
        //
        int mutability = getProperty(propertyQName).getMutability();

        return mutability == PropertyMetadata.APPENDABLE || 
               mutability == PropertyMetadata.MUTABLE;
    }

    public boolean canUpdate(QName propertyQName)
    {
        //
        // update requires write permission and full mutability (AND
        // isReadOnlyExternal() == false)
        //
        PropertyMetadata property = getProperty(propertyQName);
        return property.getMutability() == PropertyMetadata.MUTABLE;
    }

    private Collection convertToObjects(Collection values, Class type)
        throws SoapFault
    {
        Element[] asArray = new Element[values.size()];
        asArray = (Element[])values.toArray(asArray);
        
        //
        // we can return a List instead of a Set (HashSet) because 
        // the collection will be iterated over by calling code, not 
        // used for internal lookups
        //
        Object[] objects = (Object[])WsrpUtils.convertToObjects(asArray, type);
        return Arrays.asList(objects);
    }

    public String getExtendedMetadata(QName propertyQName, QName elementName)
    {
        return getProperty(propertyQName).getExtendedMetadata(elementName);
    }

    public Collection getInitialValues(QName propertyQName)
    {
        return getProperty(propertyQName).getInitialValues();
    }

    public Collection getInitialValues(QName propertyQName, Class javaType)
        throws SoapFault
    {
        Collection values = getProperty(propertyQName).getInitialValues();
        return convertToObjects(values, javaType);
    }

    public QName getInterface()
    {
        return _wsdlInterface;
    }

    public String getLowerBound(QName propertyQName)
    {
        return getProperty(propertyQName).getLowerBound();
    }

    public String getName()
    {
        return _name;
    }

    private PropertyMetadata getProperty(QName propertyQName)
    {
        PropertyMetadata property = (PropertyMetadata)_propertiesByQName.get(propertyQName);

        if (property == null)
        {
            Object[] filler = { propertyQName };
            throw new RuntimeException(_MESSAGES.get("UndefinedProperty", filler));
        }

        return property;
    }

    public Collection getPropertyNames()
    {
        return Collections.unmodifiableSet(_propertiesByQName.keySet());
    }

    public Collection getStaticValues(QName propertyQName)
    {
        return getProperty(propertyQName).getStaticValues();
    }

    public Collection getStaticValues(QName propertyQName, Class javaType)
        throws SoapFault
    {
        Collection values = getProperty(propertyQName).getStaticValues();
        return convertToObjects(values, javaType);
    }

    public String getUpperBound(QName propertyQName)
    {
        return getProperty(propertyQName).getUpperBound();
    }

    public Collection getValidValues(QName propertyQName)
    {
        return getProperty(propertyQName).getValidValues();
    }

    public Collection getValidValues(QName propertyQName, Class javaType)
        throws SoapFault
    {
        Collection values = getProperty(propertyQName).getValidValues();
        return convertToObjects(values, javaType);
    }

    public String getWsdlLocation()
    {
        return _wsdlLocation;
    }

    public boolean hasProperty(QName propertyQName)
    {
        return _propertiesByQName.containsKey(propertyQName);
    }

    public boolean isInitialValue(QName propertyQName, Object value)
    {
        return getProperty(propertyQName).isInitialValue(value);
    }

    public boolean isReadOnlyExternal(QName propertyQName)
    {
        PropertyMetadata property = getProperty(propertyQName);
        return property.getModifiability() == PropertyMetadata.READ_ONLY;
    }

    public boolean isStaticValue(QName propertyQName, Object value)
    {
        return getProperty(propertyQName).isStaticValue(value);
    }

    public boolean isValidValue(QName propertyQName, Object value)
    {
        return getProperty(propertyQName).isValidValue(value);
    }

    public Element toXML()
    {
        return toXML(XmlUtils.EMPTY_DOC);
    }

    public Element toXML(Document doc)
    {
        Element root = XmlUtils.createElement(doc, WsrmdConstants.DESCRIPTOR_QNAME);
        
        root.setAttribute(WsrmdConstants.NAME, getName());
        root.setAttribute(WsrmdConstants.INTERFACE, XmlUtils.toString(getInterface()));
        
        //
        // HACK: the namespace that is the first token of this attribute value 
        //       is incorrect - it should be the target NS of the WSDL file, 
        //       but I don't want to re-load the WSDL file just to get that 
        //       value, and I don't think it's going to be used
        //
        root.setAttribute(WsrmdConstants.WSDL_LOCATION, WsrmdConstants.NAMESPACE_URI + ' ' + getWsdlLocation());
        
        Iterator i = getPropertyNames().iterator();
        
        while (i.hasNext())
        {
            PropertyMetadata property = getProperty((QName)i.next());
            Element propertyXML = property.toXML(doc);
            root.appendChild(propertyXML);
        }
        
        return root;
    }
}
